import 'package:source_gen/source_gen.dart';
import 'package:build/build.dart';
import 'package:code_builder/code_builder.dart';
import 'package:dart_style/dart_style.dart';
import 'package:hyper/src/dom_elements.dart';

const reservedWords = [
  'abstract',
  'dynamic',
  'implements',
  'show',
  'as',
  'else',
  'import',
  'static',
  'assert',
  'enum',
  'in',
  'super',
  'async',
  'export',
  'interface',
  'switch',
  'await',
  'extends',
  'is',
  'sync',
  'break',
  'external',
  'library',
  'this',
  'case',
  'factory',
  'mixin',
  'throw',
  'catch',
  'false',
  'new',
  'true',
  'class',
  'final',
  'null',
  'try',
  'const',
  'finally',
  'on',
  'typedef',
  'continue',
  'for',
  'operator',
  'var',
  'covariant',
  'Function',
  'part',
  'void',
  'default',
  'get',
  'rethrow',
  'while',
  'deferred',
  'hide',
  'return',
  'with',
  'do',
  'if',
  'set',
  'yield',
];

String toCamelCase(String input) {
  return input.replaceAllMapped(new RegExp(r'-([a-z])'), (match) {
    return match.group(1).toUpperCase();
  });
}

final _dartfmt = DartFormatter();

const linterRules = '// ignore_for_file: non_constant_identifier_names\n';

// Creates a `<a>` element.
// Element a({Map<String, Object> attrs, List<Element> children}) =>
//     Element("a", attrs: attrs, children: children);

class HyperElementGenerator extends Generator {
  @override
  String generate(LibraryReader library, _) {
    final methods = typedElements.map((tag) {
      final parameters = [
        Parameter((b) => b
          ..type = TypeReference((b) => b
            ..symbol = 'Map'
            ..types.addAll([
              refer('String'),
              refer('String'),
            ]))
          ..name = 'attrs'
          ..named = true),
      ];
      final arguments = {'attrs': refer('attrs')};

      if (!tag.isVoidElement) {
        parameters.add(Parameter((b) => b
          ..type = TypeReference((b) => b
            ..symbol = 'List'
            ..types.addAll([
              refer('Element', 'package:hyper/hyper.dart'),
            ]))
          ..name = 'children'
          ..named = true));
        arguments['children'] = refer('children');
      }

      return Method((b) => b
        ..docs.addAll(
          ['/// Create a `<${tag.tag}>` element.'],
        )
        ..returns = refer('Element', 'package:hyper/hyper.dart')
        ..name = reservedWords.contains(tag.tag) ? '${tag.tag}_' : tag.tag
        ..optionalParameters.addAll(parameters)
        ..lambda = true
        ..body = refer('Element', 'package:hyper/hyper.dart').call(
          [literal(tag.tag)],
          arguments,
        ).statement);
    });

    return _dartfmt.format(
      linterRules + methods.map((m) => m.accept(DartEmitter())).join('\n'),
    );
  }
}

Builder hyperElementBuilder(BuilderOptions _) {
  return SharedPartBuilder([
    HyperElementGenerator(),
  ], 'hyper');
}

class HyperStrictElementGenerator extends Generator {
  @override
  String generate(LibraryReader library, _) {
    final methods = typedElements.map((tag) {
      final parameters = tag.attributes.map((a) {
        return Parameter((b) => b
          ..type = TypeReference((b) => b..symbol = 'String')
          ..name = reservedWords.contains(a.name)
              ? '${toCamelCase(a.name)}_'
              : toCamelCase(a.name)
          ..named = true
          ..docs.addAll(
            ['/// ${a.description}'],
          ));
      }).toList();

      parameters.add(Parameter((b) => b
        ..type = TypeReference((b) => b
          ..symbol = 'Map'
          ..types.addAll([
            refer('String'),
            refer('String'),
          ]))
        ..name = 'attrs'
        ..named = true));

      final arguments = {
        'attrs': refer('_attrs'),
      };

      if (!tag.isVoidElement) {
        parameters.add(Parameter((b) => b
          ..type = TypeReference((b) => b
            ..symbol = 'List'
            ..types.addAll([
              refer('Element', 'package:hyper/hyper.dart'),
            ]))
          ..name = 'children'
          ..named = true));

        arguments['children'] = refer('children');
      }

      final Map<String, Expression> attributeMap = {};
      tag.attributes.forEach((a) {
        attributeMap[a.name] = refer(reservedWords.contains(a.name)
            ? '${toCamelCase(a.name)}_'
            : toCamelCase(a.name));
      });

      return Method((b) => b
        ..docs.addAll(
          ['/// Create a `<${tag.tag}>` element.'],
        )
        ..returns = refer('Element', 'package:hyper/hyper.dart')
        ..name = reservedWords.contains(tag.tag) ? '${tag.tag}_' : tag.tag
        ..optionalParameters.addAll(parameters)
        ..body = Block.of([
          literalMap(attributeMap).assignVar('_attrs').statement,
          const Code('if (attrs != null) {'),
          refer('_attrs').property('addAll').call([refer('attrs')]).statement,
          const Code('}'),
          refer('_attrs').property('removeWhere').call([
            CodeExpression(const Code('(k, v) => v == null')),
          ]).statement,
          refer('Element', 'package:hyper/hyper.dart')
              .call([literal(tag.tag)], arguments)
              .returned
              .statement,
        ]));
    });

    return _dartfmt.format(
      linterRules + methods.map((m) => m.accept(DartEmitter())).join('\n'),
    );
  }
}

Builder hyperStrictElementBuilder(BuilderOptions _) {
  return SharedPartBuilder([
    HyperStrictElementGenerator(),
  ], 'hyper');
}
